/*  Canon Bubble Jet Print Filter.
 *  Copyright CANON INC. 2001
 *  All Rights Reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include	<signal.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<fcntl.h>
#include	<sys/wait.h>

#include	"lmsmsock.h"
#include	"lm.h"

int  sem_id = 0;			/* printer access semapho id */
int  filter_exist = 1;			/* parent process is bjfilter(flag) */
int  device_type;			/* Device type (USB/LP/PARPORT......) */
char *stamon_arg[ARGV_MAX];		/* arg space */

static int   data_end  = 0;             /* print data outputed flag */
static int   do_cancel = 0;		/* print cancel flag */
static pid_t stmon_pid = 0;		/* status monitor process id */
static pid_t data_pid  = 0;		/* data output process id */
static pid_t stat_pid  = 0;		/* status get process id */

volatile static int int_sig   = 0;	/* SIGINT flag */
volatile static int term_sig  = 0;	/* SIGTERM flag */
volatile static int user1_sig = 0;	/* print data end flag */
volatile static int user2_sig = 0;	/* SM(status monitor) show flag */

int data_output_process(int);
int status_get_process(int);
int lm_create_server_socket(void);
int lm_wait_client_connect(int);
int write_printer_command(int, char *, int);

static int print_normal(void);
static int lm_main_fork(void);
static int connect_socket(int *, int *);
static int lm_main_process(int);
static int check_signal(void);
static int check_signal_arrive(void);
static int do_prn_com(int, lmsm_socket);
static int do_prn_cancel(void);

/* signal handlers */
void sighand_int(int sig)
{
	/* SIGINT handler */
	int_sig = 1;
}

void sighand_term(int sig)
{
	/* SIGTERM handler */
	term_sig = 1;
}

void sighand_user1(int sig)
{
	/* SIGUSR1 handler */
	user1_sig = 1;
}

void sighand_user2(int sig)
{
	/* SIGUSR2 handler */
	user2_sig = 1;
}

int main(int ac, char **arv)
{
	int error;

#ifdef DEBUG
	/* logfile=/var/log/log.txt */
	create_log();
#endif

	/* setup signal handlers */
	signal_setup(SIGTERM, sighand_term);
	signal_setup(SIGUSR1, sighand_user1);
	signal_setup(SIGUSR2, sighand_user2);
	signal_setup(SIGINT,  sighand_int);

	/* printer DEVICE ID and command line option check */
	if((error = get_printer_devid()) < 0 || check_arg(ac,arv) < 0){

		/*
		   Not Canon printer
			or
		   "--gui" option not found
		*/
#ifdef DEBUG
		write_log("Now normal printing ......\n");
#endif
		print_normal();
		exit(0);
	}

	/* create semapho and other ..... */
	if((error = lm_init(PRNT_PATH))){
		exit(0);
	}

#ifdef DEBUG
	if(filter_exist)
		write_log("Parent process is bjfilter\n");
	else
		write_log("Parent process is unknown\n");
#endif

	/* SM_process/print_process/status_process start */
	lm_main_fork();

#ifdef DEBUG
	write_log("do delete semapho\n");
#endif
	/* delete semapho */
	remove_sem(sem_id);
	/* free memory (status monitor argv string) */
	free_arg();

#ifdef DEBUG
	write_log("LM end \n");
#endif
	exit(0);
}

static int lm_main_fork()
{
	int child_sm;
	int child_stat;
	int child_data;

	if((stmon_pid = fork())){
		int server_fd  = 0;
		int sm_sock_fd = 0;

		/* create socket and connect LM<->SM */
		if(connect_socket(&server_fd, &sm_sock_fd) < 0){
			/* socket error */

			if(server_fd > 0)
				close(server_fd);
			if(sm_sock_fd > 0)
				close(sm_sock_fd);

			kill(stmon_pid, SIGTERM);
			goto lm_exit;
		}

		if((stat_pid = fork())){
			if((data_pid = fork())){
				/* LM main process */

				/* exec lgmon main process */
				lm_main_process(sm_sock_fd);
				close(sm_sock_fd);
				close(server_fd);
			}
			else
				/* print data output process */
				data_output_process(sm_sock_fd); /* never return */
		}
		else
			/* printer status get process */
			status_get_process(sm_sock_fd); /* never return */
	}
	else{
		/* status monitor process */
		if(execv(STMON_PATH, stamon_arg) < 0){
#ifdef DEBUG
			write_log("SM execv error\n");
#endif
		}
		exit(0); /* never arrive this point */
	}

lm_exit:
	/* wait child process end */
	if(data_pid)
		wait(&child_data);
	if(stat_pid)
		wait(&child_stat);
	if(stmon_pid)
		wait(&child_sm);

	if(do_cancel)
		do_printer_reset(1);

	return 0;
}

static int connect_socket(int *sv, int *sk)
{
	int v,k;
	int i;

	/* create socket */
	for(i = 5; i > 0; i--){
		v = lm_create_server_socket();
		if(v >= 0)
			break; /* SUCCESS */
		else{
			if(check_signal() == -1){
#ifdef DEBUG
				write_log("socket create error(signal detected)\n");
#endif
				break;
			}
		}
#ifdef DEBUG
		write_log("socket create error(no signal and retry)\n");
#endif
	}
	if(v < 0){
		/* create error */
		*sv = -1;
		*sk = -1;
		return -1;
	}

	*sv = v;

	/* connect wait */
	for(i = 5; i > 0; i--){
#ifdef DEBUG
		write_log("start accept (LM main)\n");
#endif
		k = lm_wait_client_connect(v);
		if(k >= 0)
			break; /* SUCCESS */
		else{
			if(check_signal() == -1){
#ifdef DEBUG
				write_log("socket connect error(signal detected)\n");
#endif
				break;
			}
		}
#ifdef DEBUG
		write_log("socket connect error(no signal and retry)\n");
#endif
	}
	if(k < 0){
		/* connect error */
		*sk = -1;
		return -1;
	}

	*sk = k;

	return 0;
}

static int check_signal()
{
	/* SIGTERM/SIGCHILD/SIGINT arrive? */
	if(term_sig || int_sig)
		return -1;

	return 0;
}

static int print_normal()
{
	int error = 0;
	int r_size;
	int w_size;
	char *buf;
	char *ptr;

	buf = (char *)malloc(MAX_DATA);

	signal_setup(SIGTERM, sighand_term);

	while((r_size = read(DATA_PATH, buf, MAX_DATA)) > 0){
		ptr = buf;
		for(;r_size > 0;r_size -= w_size){
			w_size = write(PRNT_PATH, ptr, r_size);
			if(w_size < 0){
				/* write error */
				error = -1;
				goto print_normal_exit;
			}
			ptr += w_size;
		}
	}

print_normal_exit:
	if(term_sig)
		do_printer_reset(1);

	free(buf);
	return 0;
}

static int lm_main_process(int sm_sock_fd)
{
	int error = 0;
	int prn_end = 0;
	lmsm_socket sm_data;
	int first_start = 0;

#ifdef DEBUG
	write_log("into lm_main\n");
#endif

	sleep(1); /* wait child process setup */

	/*
	  LM<->SM socket read loop start.
	  signal-infomation
		SIGTERM ---- program end.
		SIGUSR1 ---- print data end.
		SIGUSR2 ---- status monitor show request.
		SIGINT  ---- program end.
	*/
	while(!prn_end){
		/* socket read */
		error = read(sm_sock_fd, &sm_data, sizeof(sm_data));

		prn_end = check_signal_arrive();

		if(error < 0)
			continue; /* socket read error */

		switch(sm_data.command){
			case LM_PRN_COM:
				/* output printer command */
				prn_end = do_prn_com(sm_sock_fd, sm_data);
				break;
			case LM_PRN_CANCEL:
				prn_end = do_prn_cancel();
				break;
			case LM_PRN_START:
#ifdef DEBUG
				write_log("SM message LM_PRN_START\n");
#endif
				if(first_start != 0)
					break;

				/* printing start!! */
				kill(data_pid, SIGUSR2); /* message to data_output_process */
				first_start = 1; /* next START is ignore */
				break;
			case LM_PRN_END:
#ifdef DEBUG
				write_log("SM message LM_PRN_END\n");
#endif
				/* end of all job's */
				/* print end(last page ejected from printer) */
				do_cancel = 1;
				prn_end = 1;
				break;
			default:
				break;
		}
	}

	/* print end */

	/* child process end */
	if(data_pid)
		kill(data_pid, SIGTERM);
	if(stat_pid)
		kill(stat_pid, SIGTERM);
	if(stmon_pid)
		kill(stmon_pid, SIGTERM);

	return 0;
}

static int check_signal_arrive(void)
{
#ifdef DEBUG
	if(term_sig)
		write_log("SIGTERM receive(LM main)\n");
	if(user1_sig)
		write_log("SIGUSR1 receive(LM main)\n");
	if(user2_sig)
		write_log("SIGUSR2 receive(LM main)\n");
	if(int_sig)
		write_log("SIGINT receive(LM main)\n");
#endif

	/* print data end from bjfilter? */
	if(user1_sig){
		/* message to data_output_process (print data is end) */
		kill(data_pid, SIGUSR1);
		user1_sig = 0;
		data_end = 1; /* print data output is end */
		return 0;
	}

	/* show SM ? */
	if(user2_sig){
		/* status monitor show signal send */
		kill(stmon_pid, SIGUSR1);
		user2_sig = 0;
		return 0;
	}

	/* SIGINT from lpd ? */
	if(int_sig){
		do_cancel = 1; /* printer reset */
		return 1; /* program end */
	}

	/* SIGTERM */
	if(term_sig)
		return 1; /* program end */

	/* no signals. */
	return 0;
}

static int do_prn_com(int sm_sock_fd, lmsm_socket sm_data)
{
	int error;
	Prcom_st ptr;

	ptr = &(sm_data.prn_data.com);
#ifdef DEBUG
	write_log("SM message LM_PRN_COM\n");
	write_log(ptr->com_buffer+9);
	write_log("\n");
#endif

	while(1){
		get_printer_sem(sem_id);     /* lock printer */
		error = write_printer_command(PRNT_PATH, ptr->com_buffer, ptr->com_size);
		release_printer_sem(sem_id); /* release printer */
		if(error){
			if(check_signal_arrive() == 1)
				return 1; /* SIGINT/SIGTERM detect(abort) */
#ifdef DEBUG
			write_log("PRN_COM error(retry)\n");
#endif
			sleep(2); /* retry */
		}
		else
			break; /* command success */
	}
#ifdef DEBUG
	write_log("PRN_COM success\n");
#endif

	get_printer_sem(sem_id); /* lock printer */
	status_to_sm(sm_sock_fd, PRNT_PATH);
	release_printer_sem(sem_id); /* release printer */

	return 0;
}

static int do_prn_cancel(void)
{
#ifdef DEBUG
	write_log("SM message LM_PRN_CANCEL\n");
#endif
	/*
	   print data already outputed to printer.
			or
	   bjfilter is not parent process. ----> force print end.
	*/
	if(data_end == 1 || filter_exist == 0){
		do_cancel = 1;
		return 1;
	}

	/* message to bjfilter (lgmon will soon end) */
	user1_sig = 0;
	kill(getppid(), SIGUSR1);

	/* wait until SIGUSR1(bjfilter ACK) receive */
	while(1){
		/* wait signal from bjfilter */
		if(user1_sig)
			break; /* ACK receive. */

#ifdef DEBUG
		write_log("wait SIGUSR1 from bjfilter\n");
#endif
		pause();
	}
#ifdef DEBUG
	write_log("receive SIGUSR1\n");
#endif
	do_cancel = 1;
	return 1;
}

