/*  Canon Bubble Jet Print Filter.
 *  Copyright CANON INC. 2001
 *  All Rights Reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>

#include <gtk/gtk.h>

#include "interface.h"
#include "support.h"
#include "callbacks.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <netdb.h>
#include <string.h>
#include <signal.h>

#include "lmsmsock.h"
#include "smstat.h"

#define  LM_COM_PORT	19999

extern	void print_sts_recv();
extern	int 	SM_ComCreate();
extern	void	sighand_term();
extern	void	sighand_usr1();
extern	gint stsmon_show_check();

char 	*check_arg(int ac,char **argv);
int  	check_noparent(int ac,char **argv);
int  	check_model_s(char *);
void	SetGtkResourceDefault();
void	draw_init(GtkWidget *);

#ifdef DEBUG
void create_log();
void write_log(char *str);
#endif

/*-------------------------------------------
 * 	GLOBAL 
 *-------------------------------------------*/
int		sock_tag;	// socket tag
int		sock_path;	// socket path
lmsm_socket	lmsm_s;
int		enable_cancel;	// enable cancel   button	

GdkPixmap	*pixmap;	// drawing area
GdkPixmap	*g_out;		// Ink_Out Fill Pixmap
GdkPixmap	*g_low;		// Ink_low Fill Pixmap
GdkBitmap	*mask_out;	// Ink Out Pixmap Mask
GdkPixmap	*g_out_bk;	// Ink_Out Fill Pixmap
GdkPixmap	*g_low_bk;	// Ink_low Fill Pixmap
GdkBitmap	*mask_out_bk;	// Ink Out Pixmap Mask
GdkPixmap	*g_I_el;	// cartridge edge left Pixmap
GdkPixmap	*g_I_er;	// cartridge edge right Pixmap
GdkPixmap	*g_I_pb;	// ink photo black Pixmap
GdkPixmap	*g_I_pc;	// ink photo cyan Pixmap
GdkPixmap	*g_I_pm;	// ink photo magenta Pixmap
GdkPixmap	*g_I_bk;	// ink black Pixmap
GdkPixmap	*g_I_bb;	// ink big black Pixmap
GdkPixmap	*g_I_cy;	// ink cyan Pixmap
GdkPixmap	*g_I_ma;	// ink magenta Pixmap
GdkPixmap	*g_I_ye;	// ink yellow Pixmap
GdkPixmap	*g_I_sc;	// scanner Pixmap
GdkPixmap	*g_I_sp;	// ink space for big-black Pixmap
GdkBitmap	*mask_nal;	// Mask nallow
GdkBitmap	*mask_wid;	// Mask wide (big black)
GdkBitmap	*mask_el;	// Mask for edge
GdkBitmap	*mask_er;	// Mask for edge
GdkBitmap	*mask_sp;	// Mask for space
GdkBitmap	*mask_sc;	// Mask for scanner

GdkPixmap	**ink_p[INK_CNT];
GdkBitmap	**mask_p[INK_CNT];

gchar		*sm_pixmaps_dir;
		
int		dw,dh;		// Pixmap Default dx,dy
volatile 	int	flag_term = 0;	
volatile 	int	flag_redisp = 0;
int		flag_noparent;	// 
gint		to_tag;		// timeout check tag
char		n_ptr_sub[256];	//
char		sm_gap;		// Paper thickness lever 
short		chd_kind;	//
short		model_id;	// model id for table
short		dev_type;	// Device type LP, USB

char		printing_stat;	// 1: Data Out OK
char		fault_stat;	// 1: Fault
char		wait_end;    	// 1: waiting for compelete Operator Call

struct pr_t	pr_tbl[MODEL_CNT] = {
		{"BJF850", BACK_GREEN},
		{"BJF860", BACK_GREEN},
		{"BJF870", BACK_GREEN},
		{"BJF360", BACK_GREEN},
		{"BJS600", BACK_BLACK},
		{"BJS630", BACK_BLACK},
		{"BJS6300",BACK_BLACK},
//		{"BJF660", BACK_GREEN},
};
struct pat_t 	pat_tbl[CHD_CNT2][CHD_AREA_MAX] = {
		{{I_el, 10},	// BC-50
		 {I_bk, 13},
		 {I_pc, 30},
		 {I_pm, 47},
		 {I_cy, 64},
		 {I_ma, 81},
		 {I_ye, 98},
		 {I_er,115},
		 {   0,  0},
		 {   0,  0}},
		{{I_el, 19},	// BC-33
		 {I_bb, 22},
		 {I_cy, 47},
		 {I_ma, 64},
		 {I_ye, 81},
		 {I_er, 98},
		 {   0,  0},
		 {   0,  0},
		 {   0,  0},
		 {   0,  0}},
		{{I_el, 29},	// BC-30
		 {I_sp, 32},
		 {I_bb, 45},
		 {I_sp, 70},
		 {I_er, 83},
		 {   0,  0},
		 {   0,  0},
		 {   0,  0},
		 {   0,  0},
		 {   0,  0}},
		{{I_el, 20},	// BC-34
		 {I_pb, 23},
		 {I_pc, 40},
		 {I_pm, 57},
		 {I_ye, 74},
		 {I_er, 91},
		 {   0,  0},
		 {   0,  0},
		 {   0,  0},
		 {   0,  0}},
		{{I_el,  0},	// BC-30+BC-31
		 {I_sp,  3},
		 {I_bb, 16},
		 {I_sp, 41},
		 {I_er, 54},
		 {I_el, 63},
		 {I_cy, 66},
		 {I_ma, 83},
		 {I_ye,100},
		 {I_er,117}},
		{{I_el,  0},	// BC-32+BC-31
		 {I_pb,  3},
		 {I_pc, 20},
		 {I_pm, 37},
		 {I_er, 54},
		 {I_el, 63},
		 {I_cy, 66},
		 {I_ma, 83},
		 {I_ye,100},
	 	 {I_er,117}}, 
//		{{I_sc, 13},	// Scanner+BC-31
		{{   0, 13},	// Scanner+BC-31
		 {I_el, 63},
		 {I_cy, 66},
		 {I_ma, 83},
		 {I_ye,100},
		 {I_er,117},
		 {   0,  0},
		 {   0,  0},
		 {   0,  0},
		 {   0,  0}},
//		{{I_sc, 29},	// Scanner
		{{   0, 29},	// Scanner
		 {   0,  0},
		 {   0,  0},
		 {   0,  0},
		 {   0,  0},
		 {   0,  0},
		 {   0,  0},
		 {   0,  0},
		 {   0,  0},
		 {   0,  0}},
};

#ifdef DEBUG
FILE *log_path;
#endif

/*-------------------------------------------
 * 	MAIN   
 *-------------------------------------------*/
int
main (int argc, char *argv[])
{
  GtkWidget 	*window1;
  GtkWidget 	*print_continue;
  GtkWidget 	*print_cancel;
  struct 	sockaddr_in address;
  struct	hostent	*host;
  struct	sigaction	sa;

  int	 	ret,	retry;
  char		*model_s,model_name[30];

#ifdef DEBUG
  create_log();
#endif

  bindtextdomain (PACKAGE, PACKAGE_LOCALE_DIR);
  textdomain (PACKAGE);

  /*-------------------------
     Initialize
  ---------------------------*/
  SetGtkResourceDefault();

  gtk_set_locale ();
  gtk_init (&argc, &argv);

  if( (model_s = check_arg( argc,argv )) != NULL )
  	if( (model_id = check_model_s( model_s )) >= MODEL_CNT) 
		model_s = NULL;
  flag_noparent = check_noparent( argc, argv );	
 
  enable_cancel = printing_stat = fault_stat = 0;
  wait_end = sm_gap = 0;

	ink_p[I_el-1] = &g_I_el;
	ink_p[I_er-1] = &g_I_er;
  	ink_p[I_pb-1] = &g_I_pb;
  	ink_p[I_bk-1] = &g_I_bk;
  	ink_p[I_bb-1] = &g_I_bb;
  	ink_p[I_pm-1] = &g_I_pm;
  	ink_p[I_pc-1] = &g_I_pc;
  	ink_p[I_cy-1] = &g_I_cy;
  	ink_p[I_ma-1] = &g_I_ma;
  	ink_p[I_ye-1] = &g_I_ye;
  	ink_p[I_sp-1] = &g_I_sp;
  	ink_p[I_sc-1] = &g_I_sc;

  	mask_p[I_el-1] = &mask_el;
  	mask_p[I_er-1] = &mask_er;
  	mask_p[I_pb-1] = &mask_nal;
  	mask_p[I_bk-1] = &mask_nal;
  	mask_p[I_bb-1] = &mask_wid;
  	mask_p[I_pm-1] = &mask_nal;
  	mask_p[I_pc-1] = &mask_nal;
  	mask_p[I_cy-1] = &mask_nal;
  	mask_p[I_ma-1] = &mask_nal;
  	mask_p[I_ye-1] = &mask_nal;
  	mask_p[I_sp-1] = &mask_sp;
  	mask_p[I_sc-1] = &mask_sc;

  /*-----------------------
  // Create Socket
  -------------------------*/
  sock_path = socket( AF_INET, SOCK_STREAM, 0);
  if( sock_path < 0 )	exit(1);

  address.sin_family = AF_INET;
  address.sin_port = htons(LM_COM_PORT);

  host = gethostbyname("localhost");
  if( !host )	close( sock_path );
  memcpy( &address.sin_addr, host->h_addr_list[0], sizeof(address.sin_addr));

  retry = 5;
  while( retry > 0)
  {
#ifdef DEBUG
	write_log("<SM> connect...\n");
#endif
	//Connect to LM
	ret = connect( sock_path, (struct sockaddr *)&address, sizeof(address));
	if( ret == -1 ){
		sleep(1);
//		retry--;
	}else	break;
  }
  if( retry == 0)
  {
	perror(" Can't connect!\n");
	exit(1);
  }

  /*---------------------------------
  // Printer Status Initial Setting
  ----------------------------------*/
  lmsm_s.socket_type = SMtoLM;
  lmsm_s.command     = LM_PRN_COM;
  lmsm_s.prn_data.com.com_size = SM_ComCreate( lmsm_s.prn_data.com.com_buffer,
				SM_INIT_BJF850 );
#ifdef DEBUG
	write_log("write INIT_COM\n");
#endif

  ret = write( sock_path, &lmsm_s, sizeof(lmsm_s));
  if( ret == -1 )	exit(1);

#ifdef UNUSE
  add_pixmap_directory (PACKAGE_DATA_DIR "/pixmaps");
  add_pixmap_directory (PACKAGE_SOURCE_DIR "/../pixmaps");
#endif

  sm_pixmaps_dir = g_strdup( PACKAGE_DATA_DIR "/pixmaps" );

  /*
   * The following code was added by Glade to create one of each component
   * (except popup menus), just so that you see something after building
   * the project. Delete any components that you don't want shown initially.
   */

  /*-------------------------------
     Create Window1
  -------------------------------*/

  window1 = create_window1 ();
  
  // Set Title 
	strcpy( model_name, "Canon ");
	if( model_s != NULL )
		strcat( model_name, model_s );
	else
		strcat( model_name, "BJ-Printer" );
	strcat( model_name, " Status Monitor" );
  	gtk_window_set_title( GTK_WINDOW (window1), model_name );

  // hide button
  print_continue = lookup_widget( window1, "print_continue" );
  print_cancel = lookup_widget( window1, "print_cancel"   );
  gtk_widget_set_sensitive( print_continue, FALSE );
  gtk_widget_set_sensitive( print_cancel, FALSE  );
  gtk_widget_show (window1);

/*-------------
 * Socket Check
 -------------*/
  sock_tag = gdk_input_add( sock_path,
				GDK_INPUT_READ,
				print_sts_recv,
				window1   );
  to_tag  = gtk_timeout_add( SM_SHOW_TIMEOUT_ms,
				stsmon_show_check,
				window1 );

  memset( &sa, 0, sizeof(sa));
  sa.sa_handler = sighand_term;
  if( sigaction( SIGTERM, &sa, NULL) ) perror("sigaction");
  sa.sa_handler = sighand_usr1;
  if( sigaction( SIGUSR1, &sa, NULL) ) perror("sigaction");
  
  /*-----------------
  // Printing Start
  ------------------*/
  lmsm_s.command     = LM_PRN_START;
  ret = write( sock_path, &lmsm_s, sizeof(lmsm_s));

  /*---------------------
    GTK_MAIN()
  ---------------------*/
  gtk_main ();

  gtk_timeout_remove( to_tag );
  gdk_input_remove( sock_tag );
  if( sock_path != -1 )  close( sock_path );

  exit (0);
}


char *check_arg( int	ac, char **argv )
{
  int	i;

  if( ac <= 1 ) return NULL;
  for( i=0; i<ac; i++){
	if(!(strcmp(argv[i],"--model")) )
		return argv[i+1];
  }
  return NULL;
}

int check_noparent( int	ac, char **argv )
{
  int	i;

  if( ac <= 1 ) return 0;
  for( i=0;i<ac;i++){
	if(!(strcmp(&(argv[i][0]),"--noparent")) )
		return 1;
  }
  return 0;
}

int check_model_s( char *model_s )
{
	int	i=0;

	for( i=0; i<MODEL_CNT; i++ )
	{
		if( !(strcmp( model_s, pr_tbl[i].pr_name) ))
			break;
	}
	return i;
}

void SetGtkResourceDefault()
{
	gchar* home_dir = NULL;
	gchar* rc_path = NULL;

	home_dir = g_get_home_dir();
	rc_path = g_strdup_printf("%s/.gtkrc",home_dir);
	gtk_rc_add_default_file(rc_path);
	g_free(rc_path);
}

#ifdef DEBUG
void
create_log()
{
	log_path = fopen( "/var/log/smlog.txt","a+");
//	if( log_path == 0 )	exit(1);
	
	write_log( "*** SM log *** \n");
}
void write_log( char *str )
{
	if( log_path == 0 )	return;
	fwrite( str, 1, strlen(str),log_path);
	fflush( log_path );
}
#endif
